/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: NetBuffer.hpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#pragma once

#include "comm.hpp"
#include "common.hpp"
#include <boost/asio/buffer.hpp>
#include <deque>


namespace EPS
{
	/** NetBuffer objects manage blocks of buffer bytes used for either reading from, and writing to sockets.
	 * Included in this class is the ability to manage encryption and decryption of %EPS network data.  When a buffer
	 * is either encrypted or decrypted, a link to the original NetBuffer object is kept so the original data can
	 * easily be referenced without having to re-encrypt or re-decrypt the data should it be necessary to get access
	 * to the original bytes.
	 */
	class NetBuffer
	{
		public:

			/// Enumerator to describe the contents of this buffer.
			enum EType
			{
				kInvalid			= 0,	///< unused buffer
				kEncrypted			= 1,	///< encrypted buffer
				kPartiallyDecrypted	= 2,	///< first 16 bytes (header) is decrypted
				kFullyDecrypted		= 3,	///< decrypted buffer
				kMax
			};

			/// Constructor.
			NetBuffer(const EPS::NetBuffer::EType type = EPS::NetBuffer::EType::kInvalid);

			/// Constructor.  Calls @ref clone().
			NetBuffer(const NetBuffer &nb);

			/// Destructor.
			~NetBuffer(void);

			/// Clear out a buffer so it can be re-used. @{
			void clear(const bool also_clear_related_buffer);
			void clear(const EPS::NetBuffer::EType type = EPS::NetBuffer::EType::kInvalid);
			/// @}

			/// Return @p TRUE if the buffer is encrypted.
			bool is_encrypted(void)
			{
				return	this->buffer_type == EPS::NetBuffer::EType::kEncrypted;
			}

			/// Return @p TRUE if the buffer is partially decrypted.
			bool is_partially_decrypted(void)
			{
				return	this->buffer_type == EPS::NetBuffer::EType::kPartiallyDecrypted;
			}

			/// Return @p TRUE if the buffer is fully decrypted.
			bool is_fully_decrypted(void)
			{
				return	this->buffer_type == EPS::NetBuffer::EType::kFullyDecrypted;
			}

			/// Return @p TRUE if the buffer is either partially or fully decrypted.
			bool is_decrypted(void)
			{
				return	this->buffer_type == EPS::NetBuffer::EType::kPartiallyDecrypted	||
						this->buffer_type == EPS::NetBuffer::EType::kFullyDecrypted		;
			}

			/// Convert the buffer to the specified type.  @{
			NetBuffer *get					(const EPS::NetBuffer::EType type);
			NetBuffer *get_encrypted		(void) { return this->get(EPS::NetBuffer::EType::kEncrypted);			}
			NetBuffer *get_decrypted		(void) { return this->get(EPS::NetBuffer::EType::kFullyDecrypted);		}
			NetBuffer *get_decrypted_header	(void) { return this->get(EPS::NetBuffer::EType::kPartiallyDecrypted);	}
			/// @}

			/// Check that both the original lenght and encryption lengths are valid.
			bool lengths_are_valid(void);

			/// Convert the type to a text string. @{
			std::string name(void) const;
			std::string name(const EPS::NetBuffer::EType type) const;

			/// Return a pointer to the data. @{
			operator uint8_t*					(void)			{ return buffer.data(); }
			const uint8_t	*data				(void) const	{ return buffer.data(); }
			uint8_t			*data				(void)			{ return buffer.data(); }
			uint8_t			*data_after_header	(void)			{ return data() + sizeof(EPS::CommHeader1); }
			/// @}

			/// Get the decrypted header pointer.
			EPS::CommHeader1 *get_header(void) { return reinterpret_cast<EPS::CommHeader1*>( this->get_decrypted_header()->data() ); }

			/// Create and setup the related buffer pointer.
			EPS::NetBuffer *create_related_buffer(const EPS::NetBuffer::EType type);

			/// Make a copy of the NetBuffer (ignoring the related pointer).
			void clone(const EPS::NetBuffer &nb);

			size_t				length;			///< number of usable bytes in the buffer
			EPS::NetBuffer *	related;		///< a buffer which is related to this one through either encryption or decryption
			EType				buffer_type;	///< @see @ref buffer

			///< Buffer used to read/write to sockets, as well as encrypt/decrypt packet payloads. @{
			typedef std::array<uint8_t, EPS::network_buffer_size> Buffer;
			Buffer				buffer;			///< @see @ref buffer_type
			/// @}

			boost::asio::mutable_buffers_1	buffer_1st;		///< buffer used by asio; references buffer[0]
			boost::asio::mutable_buffers_1	buffer_2nd;		///< buffer used by asio; references buffer[16] (aka EPS::cipher_block_size, to bypass the common header)

	};

	/// Asio async_write() calls cannot overlap, so store pending buffers to send in a deque.
	typedef std::deque<EPS::NetBuffer*> NBDeque;
}
